//
// nono
// Copyright (C) 2020 nono project
// Licensed under nono-license.txt
//

//
// ネットワークのホストデバイス
//

// ログ名
//       Ethernet    HostDevice       NetDriver
//         |          |                |
//         v          v                v
//        Lance   -> HostNetDevice -> NetDriver***
// 表示名 (Lance)    (HostNet0)      (HostNet0.***)
// 識別名 "lance"    "hostnet0"      "hostnet0"
//
// HostNet は、特に NetDriver 派生クラス選択時はフォールバックも含めて
// どれが選択されるか複雑なこともあるので、ログ表示名に自分のドライバ名
// も表示したい。

#include "hostnet.h"
#include "config.h"
#include "monitor.h"
#include "netdriver_none.h"
#if defined(HAVE_HOSTNET_AFPACKET)
#include "netdriver_afpacket.h"
#endif
#if defined(HAVE_HOSTNET_BPF)
#include "netdriver_bpf.h"
#endif
#if defined(HAVE_HOSTNET_TAP)
#include "netdriver_tap.h"
#endif
#include <zlib.h>

// コンパイル済みのドライバ名一覧を返す (MainApp から呼ばれる)
/*static*/ std::vector<std::string>
HostNetDevice::GetDrivers()
{
	std::vector<std::string> list;

#if defined(HAVE_HOSTNET_AFPACKET)
	list.emplace_back("afpacket");
#endif
#if defined(HAVE_HOSTNET_BPF)
	list.emplace_back("bpf");
#endif
#if defined(HAVE_HOSTNET_TAP)
	list.emplace_back("tap");
#endif

	return list;
}


// コンストラクタ
HostNetDevice::HostNetDevice(Device *parent_, uint n)
	: inherited(parent_, OBJ_HOSTNET(n))
{
	monitor = gMonitorManager->Regist(ID_MONITOR_HOSTNET(n), this);
	monitor->func = ToMonitorCallback(&HostNetDevice::MonitorUpdate);
	monitor->SetSize(48, 23);
}

// デストラクタ
HostNetDevice::~HostNetDevice()
{
}

// ログレベル設定
void
HostNetDevice::SetLogLevel(int loglevel_)
{
	inherited::SetLogLevel(loglevel_);

	if ((bool)driver) {
		driver->SetLogLevel(loglevel_);
	}
}

// 初期化
bool
HostNetDevice::Init()
{
	if (inherited::Init() == false) {
		return false;
	}

	if (SelectDriver() == false) {
		return false;
	}

	return true;
}

// ドライバ(再)選択
bool
HostNetDevice::SelectDriver()
{
	int n = GetId() - OBJ_HOSTNET0;
	const std::string key_driver = string_format("hostnet%d-driver", n);
	const std::string key_fallback = string_format("hostnet%d-fallback", n);
	const ConfigItem& item = gConfig->Find(key_driver);
	const std::string& type = item.AsString();
	bool fallback = gConfig->Find(key_fallback).AsInt();

	driver.reset();

	if (type != "none") {
		if (type == "tap") {
#if defined(HAVE_HOSTNET_TAP)
			CreateTap();
#else
			item.Err("Hostnet driver %s not compiled", type.c_str());
#endif

		} else if (type == "bpf") {
#if defined(HAVE_HOSTNET_BPF)
			CreateBPF();
#else
			item.Err("Hostnet driver %s not compiled", type.c_str());
#endif

		} else if (type == "afpacket") {
#if defined(HAVE_HOSTNET_AFPACKET)
			CreateAFPacket();
#else
			item.Err("Hostnet driver %s not compiled", type.c_str());
#endif

		} else if (type == "auto") {
			// auto なら Tap -> { BPF, AFPacket } の順で試す。
			//
			// Tap は事前の準備が面倒だが制約がない。
			// BPF, AFPacket (この2つは OS の違いだけでほぼ同じもの) は
			// 準備に手間がかからないが自ホストと通信できない制約がある。
			// そのためどの環境でも Tap が使えれば Tap を、そうでなければ
			// BPF, AFPacket を使う、という順にしておく。
			bool r;

			r = CreateTap();
			if (r == false) {
				r = CreateAFPacket();
			}
			if (r == false) {
				r = CreateBPF();
			}

		} else {
			// 知らないドライバ種別
			item.Err();
		}

		if ((bool)driver == false) {
			putmsg(1, "hostnet-fallback=%d", fallback ? 1 : 0);

			if (fallback == false) {
				// フォールバックしないならエラー終了
				errmsg = "No host network driver found";
				putmsg(1, "%s", errmsg.c_str());
				warnx("%s (See details with options -C -Lhostnet=1)",
					errmsg.c_str());

				return false;
			}
		}
	}

	if ((bool)driver == false) {
		CreateNone();
	}
	assert((bool)driver);

	// ドライバ名は Capitalize だがログはほぼ小文字なので雰囲気を揃える…
	putmsg(1, "selected host driver: %s",
		string_tolower(driver->GetDriverName()).c_str());

	return true;
}

// None ドライバを生成する
bool
HostNetDevice::CreateNone()
{
	driver.reset(new NetDriverNone(this));
	if ((bool)driver) {
		if (driver->InitDriver()) {
			// 成功
			return true;
		}
		errmsg = driver->errmsg;
		driver.reset();
	}
	return false;
}

// Tap ドライバを生成する
bool
HostNetDevice::CreateTap()
{
#if defined(HAVE_HOSTNET_TAP)
	int n = GetId() - OBJ_HOSTNET0;
	const std::string keyname = string_format("hostnet%d-tap-devpath", n);
	const ConfigItem& item = gConfig->Find(keyname);
	const std::string& devpath = item.AsString();

	driver.reset(new NetDriverTap(this, devpath));
	if ((bool)driver) {
		if (driver->InitDriver()) {
			// 成功
			return true;
		}
		errmsg = driver->errmsg;
		driver.reset();
	}
#endif
	return false;
}

// bpf ドライバを生成する
bool
HostNetDevice::CreateBPF()
{
#if defined(HAVE_HOSTNET_BPF)
	int n = GetId() - OBJ_HOSTNET0;
	const std::string keyname = string_format("hostnet%d-bpf-ifname", n);
	const ConfigItem& item = gConfig->Find(keyname);
	const std::string& ifname = item.AsString();

	driver.reset(new NetDriverBPF(this, ifname));
	if ((bool)driver) {
		if (driver->InitDriver()) {
			// 成功
			return true;
		}
		errmsg = driver->errmsg;
		driver.reset();
	}
#endif
	return false;
}

// AFPacket ドライバを生成する
bool
HostNetDevice::CreateAFPacket()
{
#if defined(HAVE_HOSTNET_AFPACKET)
	int n = GetId() - OBJ_HOSTNET0;
	const std::string keyname = string_format("hostnet%d-afpacket-ifname", n);
	const ConfigItem& item = gConfig->Find(keyname);
	const std::string& ifname = item.AsString();

	driver.reset(new NetDriverAFPacket(this, ifname));
	if ((bool)driver) {
		if (driver->InitDriver()) {
			// 成功
			return true;
		}
		errmsg = driver->errmsg;
		driver.reset();
	}
#endif
	return false;
}

// VM からの送信 (VM スレッドで呼ばれる)
bool
HostNetDevice::Tx(const NetPacket& packet)
{
	// 長さ 0 のパケットは破棄する。バグでもないと通常は起きない。
	// runt frame はどうするか?
	if (packet.length < 1) {
		stat.txzero_pkts++;
		putlog(1, "Tx: drop empty packet");
		return true;
	}

	// 送信キューに入れて..
	if (txq.Enqueue(packet) == false) {
		stat.txqfull_pkts++;
		stat.txqfull_bytes += packet.length;
		putlog(2, "txq exhausted");
		return false;
	}

	stat.tx_pkts++;
	stat.tx_bytes += packet.length;

	// ざっくりピーク値
	stat.txq_peak = std::max((uint)txq.Length(), stat.txq_peak);

	// パイプに通知 (値はダミー)
	return WritePipe(0);
}

// rxq への投入を行う場合は true。
// VM スレッドから呼ばれる。
void
HostNetDevice::EnableRx(bool enable)
{
	rx_enable = enable;

	// 受信を停止した時はキューに残っているのを破棄する。
	// (ゲスト再起動によるデバイスリセット時とか)
	if (rx_enable == false) {
		// 統計情報に足すためだけに一旦取り出す。
		NetPacket drop;
		while (rxq.Dequeue(&drop)) {
			stat.rxdisable_pkts++;
			stat.rxdisable_bytes += drop.length;
		}
	}
}

// 受信用の自身の MAC アドレスを設定する。
void
HostNetDevice::SetMyAddr(const macaddr_t& myaddr_)
{
	myaddr = myaddr_;
}

// マルチキャストフィルタを設定する。
void
HostNetDevice::SetMCastFilter(uint64 filter_)
{
	multicast_filter = filter_;
}

// プロミスキャスモードを設定する。
void
HostNetDevice::SetPromisc(bool promisc_)
{
	promisc = promisc_;
}

// パケットをキューから取り出す。
// VM 側から呼ばれる。
bool
HostNetDevice::Rx(NetPacket *dst)
{
	if (rxq.Dequeue(dst) == false) {
		// キューが空
		return false;
	}
	stat.rx_pkts++;
	stat.rx_bytes += dst->length;
	if (__predict_false(loglevel >= 2)) {
		putlogn("Recv %u bytes", dst->length);
		if (loglevel >= 3) {
			std::string buf;
			for (uint i = 0; i < dst->length; i++) {
				if (i % 16 == 0) {
					buf += string_format("%04x:", i);
				}
				buf += string_format(" %02x", (*dst)[i]);
				if (i % 16 == 7) {
					buf += " ";
				} else if (i % 16 == 15) {
					putlogn("%s", buf.c_str());
					buf = "";
				}
			}
			if (buf.empty() == false) {
				putlogn("%s", buf.c_str());
			}
		}
	}

	return true;
}

// 外部からの読み込み (パケットを受信)。
// 戻り値はキューに投入したパケット数。
int
HostNetDevice::Read()
{
	int queued = 0;
	int left = 0;

	assert((bool)driver);

	do {
		NetPacket *p;

		p = rxq.BeginWrite();
		if (p == NULL) {
			NetPacket discard;
			left = driver->Read(&discard);
			if (left < 0) {
				break;
			}
			stat.rxqfull_pkts++;
			stat.rxqfull_bytes += discard.length;
			continue;
		}

		// driver->Read() は VM に投入すべきパケットがなければ負数を返す。
		// パケットがあれば p に書き戻して、driver 側のバッファに残っている
		// パケット数 (といっても 0 か 1以上) を返してくる。
		left = driver->Read(p);
		if (left < 0) {
			rxq.CancelWrite();
			break;
		}

		NetPacket& packet = *p;

		stat.read_pkts++;
		stat.read_bytes += packet.length;

		if (rx_enable == false) {
			rxq.CancelWrite();
			stat.rxdisable_pkts++;
			stat.rxdisable_bytes += packet.length;
			continue;
		}

		// イーサネットフレームより長いパケットはゲスト OS が期待していない。
		// ここでドロップしてみる。
		// これによりゲストでの dropping chained buffer が出なくなる。
		if (packet.length > 1518) {
			rxq.CancelWrite();
			stat.rxjumbo_pkts++;
			stat.rxjumbo_bytes += packet.length;
			continue;
		}

		// ホストカーネルから直接着信したパケット(というかフレームか)は
		// 60バイトパディングされないまま読み出せるので、ここでパディングする。
		while (packet.length < 60) {
			packet.Append(0x00);
		}

		// パケットフィルタリング
		if (__predict_true(promisc == false)) {
			static std::array<uint8, 6> bcast {
				0xff, 0xff, 0xff, 0xff, 0xff, 0xff
			};
			if (memcmp(&packet[0], &bcast[0], bcast.size()) == 0) {
				// ブロードキャストは素通し。
			} else if (__predict_false(packet[0] & 1)) {
				// マルチキャスト。
				// LANCE では LADRF レジスタ、NE2000 では MAR0-7 レジスタで
				// データシートを見る限りはロジックは同じもののようだ。

				// 宛先アドレスの CRC32 を計算する。
				// イーサネットの CRC は zlib の crc32() と同じもの。
				uint32 crc = crc32(0, Z_NULL, 0);
				crc = crc32(crc, &packet[0], 6);

				// crc の上位6ビットをとりだして検査ビット位置とする(=pos)。
				// multicast_filter の pos の位置のビットが 1 なら受信、
				// 0 なら破棄。
				// multicast_filter は64ビットで LSB が0番目、MSB が63番目
				// としたもの (NIC 側でこの通りに並べ替えて渡してある)。
				uint pos = (crc >> 26) & 0x3f;
				uint64 bit = 1U << pos;
				if ((multicast_filter & bit) == 0) {
					putlog(2, "mcast %02x:%02x:%02x:%02x:%02x:%02x pos=%d drop",
						packet[0], packet[1], packet[2],
						packet[3], packet[4], packet[5], pos);
					stat.rxmcast_pkts++;
					stat.rxmcast_bytes += packet.length;
					rxq.CancelWrite();
					continue;
				}
				putlog(2, "mcast %02x:%02x:%02x:%02x:%02x:%02x pos=%d pass",
					packet[0], packet[1], packet[2],
					packet[3], packet[4], packet[5], pos);
			} else {
				// ユニキャスト
				if (memcmp(&packet[0], &myaddr[0], 6) != 0) {
					stat.rxnotme_pkts++;
					stat.rxnotme_bytes += packet.length;
					rxq.CancelWrite();
					continue;
				}
			}
		}

		// CRC XXX 計算すること
		for (int i = 0; i < 4; i++) {
			packet.Append(0x00);
		}

		rxq.EndWrite();
		queued++;
	} while (left > 0);

	// ざっくりピーク値
	stat.rxq_peak = std::max((uint)rxq.Length(), stat.rxq_peak);

	return queued;
}

// 外部への書き出し(パケットを送信)
void
HostNetDevice::Write(uint32 data)
{
	const NetPacket *p;

	assert((bool)driver);

	// 送信キューから取り出す
	while ((p = txq.BeginRead()) != NULL) {
		const NetPacket& packet = *p;

		if (__predict_false(loglevel >= 2)) {
			putlogn("Send %u bytes", packet.length);
			if (loglevel >= 3) {
				std::string buf;
				for (uint i = 0; i < packet.length; i++) {
					if (i % 16 == 0) {
						buf += string_format("%04x:", i);
					}
					buf += string_format(" %02x", packet[i]);
					if (i % 16 == 7) {
						buf += " ";
					} else if (i % 16 == 15) {
						putlogn("%s", buf.c_str());
						buf = "";
					}
				}
				if (buf.empty() == false) {
					putlogn("%s", buf.c_str());
				}
			}
		}
		stat.write_pkts++;
		stat.write_bytes += packet.length;

		driver->Write(packet.data(), packet.length);
		txq.EndRead();
	}
}

// ドライバ名を返す
const std::string
HostNetDevice::GetDriverName() const
{
	assert((bool)driver);
	return driver->GetDriverName();
}

// モニタ
void
HostNetDevice::MonitorUpdate(Monitor *, TextScreen& screen)
{
	screen.Clear();

	screen.Print(0, 0, "Parent Device : %s", parent->GetName().c_str());
	screen.Print(0, 1, "HostNet Driver: %s", driver->GetDriverName());
	// 次の2行はドライバ依存情報
	driver->MonitorUpdateMD(screen, 2);

	// 01234567890123456
	// Write to host
	// Drop:TxQ Full
	// Drop:Rx Disabled
	// Drop:Jumbo Frame
	// Drop:Addr Filter

	int y = 5;
	screen.Print(0, y++, "%-17s%13s %17s", "<Tx>", "Packets", "Bytes");
	screen.Print(0, y++, "%-17s%13s %17s", "VM sends",
		format_number(stat.tx_pkts).c_str(),
		format_number(stat.tx_bytes).c_str());
	screen.Print(0, y++, "%-17s%13s %17s", "Write to host",
		format_number(stat.write_pkts).c_str(),
		format_number(stat.write_bytes).c_str());
	screen.Print(0, y++, "%-17s%13s %17s", "Drop:Empty Frame",
		format_number(stat.txzero_pkts).c_str(),
		"0");
	screen.Print(0, y++, "%-17s%13s %17s", "Drop:TxQ Full",
		format_number(stat.txqfull_pkts).c_str(),
		format_number(stat.txqfull_bytes).c_str());
	y++;

	screen.Print(0, y++, "%-17s%13s %17s", "<Rx>", "Packets", "Bytes");
	screen.Print(0, y++, "%-17s%13s %17s", "Read from host",
		format_number(stat.read_pkts).c_str(),
		format_number(stat.read_bytes).c_str());
	screen.Print(0, y++, "%-17s%13s %17s", "VM receives",
		format_number(stat.rx_pkts).c_str(),
		format_number(stat.rx_bytes).c_str());
	screen.Print(0, y++, "%-17s%13s %17s", "Drop:Rx Disabled",
		format_number(stat.rxdisable_pkts).c_str(),
		format_number(stat.rxdisable_bytes).c_str());
	screen.Print(0, y++, "%-17s%13s %17s", "Drop:Jumbo Frame",
		format_number(stat.rxjumbo_pkts).c_str(),
		format_number(stat.rxjumbo_bytes).c_str());
	screen.Print(0, y++, "%-17s%13s %17s", "Drop:RxQ Full",
		format_number(stat.rxqfull_pkts).c_str(),
		format_number(stat.rxqfull_bytes).c_str());
	screen.Print(0, y++, "%-17s%13s %17s", "Drop:Unicast",
		format_number(stat.rxnotme_pkts).c_str(),
		format_number(stat.rxnotme_bytes).c_str());
	screen.Print(0, y++, "%-17s%13s %17s", "Drop:Multicast",
		format_number(stat.rxmcast_pkts).c_str(),
		format_number(stat.rxmcast_bytes).c_str());
	y++;

	screen.Print(5, y++, "Capacity Peak");
	screen.Print(0, y++, "TxQ %4u/%4u %4u",
		(uint)txq.Length(), (uint)txq.Capacity(), stat.txq_peak);
	screen.Print(0, y++, "RxQ %4u/%4u %4u",
		(uint)rxq.Length(), (uint)rxq.Capacity(), stat.rxq_peak);
}
