//
// nono
// Copyright (C) 2020 nono project
// Licensed under nono-license.txt
//

//
// ネットワークのホストデバイス
//

#pragma once

#include "hostdevice.h"
#include "macaddr.h"
#include "netdriver.h"
#include "spscqueue.h"

// パケット
class NetPacket : public std::array<uint8, 1600>
{
	// ジャンボフレームでないイーサネットフレームは最大 1518バイトだが
	// このパケットバッファの長さは 1519 バイト以上必要。
	// vm/lance が 1518 バイトを越えたことを検出するのに使っている。

 public:
	NetPacket() {
		Clear();
	}

	// バッファをクリアする
	void Clear() {
		length = 0;
		Seek(0);
	}

	// パケットに1バイト追加する
	void Append(uint8 data) { (*this)[length++] = data; }

	// 読み出し開始位置を設定
	void Seek(int offset_) { offset = offset_; }

	// 読み出しの残りバイト数を返す
	int GetRemain() const { return length - offset; }

	// 1バイト読み出す。EOP なら -1 を返す
	uint32 Read() {
		if (offset < length) {
			return (*this)[offset++];
		} else {
			return (uint32)-1;
		}
	}

	// バッファ内の有効長
	uint length {};

	// 現在の読み出し位置
	int offset {};
};

class HostNetDevice : public HostDevice
{
	using inherited = HostDevice;
	using NetTxQueue = SPSCQueue<NetPacket, 32>;
	using NetRxQueue = SPSCQueue<NetPacket, 64>;

	// 統計情報
	struct stat_t {
		uint64 tx_pkts;
		uint64 tx_bytes;
		uint64 rx_pkts;
		uint64 rx_bytes;
		uint64 read_pkts;
		uint64 read_bytes;
		uint64 write_pkts;
		uint64 write_bytes;
		uint64 txzero_pkts;			// 送信長0で破棄したパケット数
		uint64 txqfull_pkts;		// キューが一杯だった回数
		uint64 txqfull_bytes;		// キューに入れられなかったバイト数
		uint64 rxdisable_pkts;		// 上位層が受信無効
		uint64 rxdisable_bytes;		// 上位層が受信無効
		uint64 rxjumbo_pkts;		// 受信したジャンボパケット数
		uint64 rxjumbo_bytes;		// 受信したジャンボパケットのバイト数
		uint64 rxqfull_pkts;		// キューが一杯だった回数
		uint64 rxqfull_bytes;		// キューに入れられなかったバイト数
		uint64 rxnotme_pkts;		// ユニキャストで破棄したパケット数
		uint64 rxnotme_bytes;		// フィルタされたバイト数
		uint64 rxmcast_pkts;		// マルチキャストで破棄したパケット数
		uint64 rxmcast_bytes;		// フィルタされたバイト数

		uint txq_peak;				// キューのおおよその最大使用量
		uint rxq_peak;				// キューのおおよその最大使用量
	};

 public:
	HostNetDevice(Device *parent_, uint n);
	~HostNetDevice() override;

	void SetLogLevel(int loglevel_) override;
	bool Init() override;

	// パケットを送信する。
	// data はプリアンブルを含まず、イーサネットフレームヘッダ(14バイト)、
	// データ(46-1500バイト) からなる。FCS(CRC) (4バイト) は含まない。
	// 送信キューに追加できれば true、出来なければ false を返す。
	bool Tx(const NetPacket& packet);

	// パケットを引き取るために VM から呼ばれる。
	// パケットが受信キューにあるときは packet にコピーして true を返す。
	// キューが空なら false を返す。
	bool Rx(NetPacket *packet);

	// 受信キューへの投入を許可/禁止する。
	void EnableRx(bool enable);

	// 受信用の自身の MAC アドレスを設定する。
	void SetMyAddr(const macaddr_t& myaddr_);

	// プロミスキャスモードを設定する。
	void SetPromisc(bool promisc_);

	// マルチキャストフィルタを設定する。
	void SetMCastFilter(uint64 filter);

	// ドライバ名を返す (ステータスパネルから呼ばれる)
	const std::string GetDriverName() const;

	// 統計情報を返す (ステータスパネルから呼ばれる)
	uint64 GetTXPkts() const { return stat.tx_pkts; }
	uint64 GetRXPkts() const { return stat.rx_pkts; }

	// コンパイル済みのドライバ名一覧を返す
	static std::vector<std::string> GetDrivers();

 private:
	bool SelectDriver();
	bool CreateNone();
	bool CreateTap();
	bool CreateBPF();
	bool CreateAFPacket();

	int Read() override;

	// 外部への書き出し(パケットを送信)
	void Write(uint32 data) override;

	DECLARE_MONITOR_CALLBACK(MonitorUpdate);

	std::unique_ptr<NetDriver> driver {};

	bool rx_enable {};
	macaddr_t myaddr {};		// 自身の MAC アドレス
	bool promisc {};			// プロミスキャスモードなら true

	// マルチキャストフィルタ。
	// 64ビットのうち、LSB 側を0番目、MSB 側を63番目となるようにしたもの。
	uint64 multicast_filter {};

	// キュー
	NetTxQueue txq {};
	NetRxQueue rxq {};

	// 統計情報
	struct stat_t stat {};

	std::string errmsg {};

	Monitor *monitor {};
};
