//
// nono
// Copyright (C) 2020 nono project
// Licensed under nono-license.txt
//

//
// MFP (MC68901)
//

#include "mfp.h"
#include "interrupt.h"
#include "keyboard.h"
#include "monitor.h"
#include "mpu.h"
#include "scheduler.h"
#include "x68kkbd.h"

// InsideOut p.135
static const busdata wait = busdata::Wait(24 * 40_nsec);

// コンストラクタ
MFPDevice::MFPDevice()
	: inherited(OBJ_MFP)
{
	monitor = gMonitorManager->Regist(ID_MONITOR_MFP, this);
	monitor->func = ToMonitorCallback(&MFPDevice::MonitorUpdate);
	monitor->SetSize(77, 26);
}

// デストラクタ
MFPDevice::~MFPDevice()
{
}

// 初期化
bool
MFPDevice::Init()
{
	if (inherited::Init() == false) {
		return false;
	}

	interrupt = GetInterruptDevice();
	keyboard = dynamic_cast<X68030Keyboard *>(GetKeyboard());

	for (int ch = 0; ch < event.size(); ch++) {
		event[ch].dev = this;
		event[ch].func = ToEventCallback(&MFPDevice::TimerCallback);
		event[ch].code = ch;
		event[ch].SetName(string_format("MFP Timer-%c", ch + 'A'));
		scheduler->RegistEvent(event[ch]);
	}

	// キーボード周りは x68kkbd.cpp のコメント参照
	key_event.func = ToEventCallback(&MFPDevice::KeyCallback);
	key_event.time = 3750_usec;
	key_event.SetName("MFP USART");
	scheduler->RegistEvent(key_event);

	return true;
}

// リセット
void
MFPDevice::ResetHard(bool poweron)
{
	// 3.3 RESET OPERATION には
	// All internal registers are cleared expect the TxDR, UDR, and TSR.
	// と書いてあるが、6.2.1 には TxDR はリセット時に $00 と書いてある。
	// 内部カウンタは影響を受けずに next_tdr だけリセットされるという
	// 意味だろうか。
	mfp.aer = 0;
	mfp.ddr = 0;
	mfp.ier.w = 0;
	mfp.ipr.w = 0;
	mfp.isr.w = 0;
	mfp.imr.w = 0;
	for (int i = 0; i < countof(mfp.timer); i++) {
		mfp.timer[i].tcr = 0;
		mfp.timer[i].next_tdr = 256;
	}
	mfp.scr = 0;
	mfp.ucr = 0;
	mfp.rsr = 0;
	// XXX TSR は変化しないがたぶんバッファはクリアされるので BE は立つ
	// ということかな。
	mfp.tsr = MFP::TSR_BE;

	// All timers are stopped.
	for (auto& ev : event) {
		scheduler->StopEvent(ev);
	}
	scheduler->StopEvent(key_event);

	// USART RX and TX are disabled.
	// SO line is placed in HighZ.
	// The interrupt channels are disabled, and
	// any pending interrupts are cleared.
	// GPIO lines are placed in HighZ.
	// timer outputs are driven low.
	// External signals are nagated.
	// VR is initialized to $00 (not $0f).
	mfp.vr_vec = 0;
	mfp.vr_s   = 0;

	// CIRQ | FMIRQ | EXPON | ALARM にしておく
	mfp.gpip = 0x6b;

	ChangeInterrupt();
}

busdata
MFPDevice::ReadPort(uint32 offset)
{
	busdata data;

	switch (offset) {
	 case MFP::GPIP:
		data = mfp.gpip;
		break;
	 case MFP::AER:
		data = mfp.aer;
		break;
	 case MFP::DDR:
		data = mfp.ddr;
		break;
	 case MFP::IERA:
		data = mfp.ier.a;
		break;
	 case MFP::IERB:
		data = mfp.ier.b;
		break;
	 case MFP::IPRA:
		data = mfp.ipr.a;
		break;
	 case MFP::IPRB:
		data = mfp.ipr.b;
		break;
	 case MFP::ISRA:
		data = mfp.isr.a;
		break;
	 case MFP::ISRB:
		data = mfp.isr.b;
		break;
	 case MFP::IMRA:
		data = mfp.imr.a;
		break;
	 case MFP::IMRB:
		data = mfp.imr.b;
		break;
	 case MFP::VR:
		data = mfp.GetVR();
		break;
	 case MFP::TACR:
		data = mfp.timer[0].tcr;
		break;
	 case MFP::TBCR:
		data = mfp.timer[1].tcr;
		break;
	 case MFP::TCDCR:
		data = mfp.GetTCDCR();
		break;
	 case MFP::TADR:
		data = GetTDR(0);
		break;
	 case MFP::TBDR:
		data = GetTDR(1);
		break;
	 case MFP::TCDR:
		data = GetTDR(2);
		break;
	 case MFP::TDDR:
		data = GetTDR(3);
		break;
	 case MFP::SCR:
		data = mfp.scr;
		break;
	 case MFP::UCR:
		data = mfp.ucr;
		break;
	 case MFP::RSR:
		data = mfp.rsr;
		break;
	 case MFP::TSR:
		data = mfp.tsr;
		break;
	 case MFP::UDR:
		// UDR からデータを読み出したら Buffer Full をクリア
		data = mfp.udr;
		ClearBF();
		break;
	 default:
		data = 0xff;
		break;
	}

	if (__predict_false(loglevel >= 3)) {
		if (offset == MFP::TCDR) {
			putlog(4, "%s -> $%02x", regname[offset], data.Data());
		} else if (offset < MFP::RegMax) {
			putlogn("%s -> $%02x", regname[offset], data.Data());
		} else {
			putlogn("$%06x -> $%02x", mpu->GetPaddr(), data.Data());
		}
	}
	data |= wait;
	data |= BusData::Size1;
	return data;
}

busdata
MFPDevice::WritePort(uint32 offset, uint32 data)
{
	switch (offset) {
	 case MFP::GPIP:
		putlog(0, "Write GPIP <- $%02x (NOT IMPLEMENTED)", data);
		mfp.gpip = data;
		break;
	 case MFP::AER:
		putlog(2, "%s <- $%02x", regname[offset], data);
		mfp.aer = data;
		break;
	 case MFP::DDR:
		if (data != 0) {
			putlog(0, "Write DDR <- $%02x (NOT IMPLEMENTED)", data);
		}
		mfp.ddr = data;
		break;
	 case MFP::IERA:
		putlog(2, "%s <- $%02x", regname[offset], data);
		SetIER(mfp.ier.a, mfp.ipr.a, data);
		break;
	 case MFP::IERB:
		putlog(2, "%s <- $%02x", regname[offset], data);
		SetIER(mfp.ier.b, mfp.ipr.b, data);
		break;
	 case MFP::IPRA:
		putlog(2, "%s <- $%02x", regname[offset], data);
		mfp.ipr.a = data;
		ChangeInterrupt();
		break;
	 case MFP::IPRB:
		putlog(2, "%s <- $%02x", regname[offset], data);
		mfp.ipr.b = data;
		ChangeInterrupt();
		break;
	 case MFP::ISRA:
		putlog(2, "%s <- $%02x", regname[offset], data);
		mfp.isr.a = data;
		break;
	 case MFP::ISRB:
		putlog(2, "%s <- $%02x", regname[offset], data);
		mfp.isr.b = data;
		break;
	 case MFP::IMRA:
		putlog(2, "%s <- $%02x", regname[offset], data);
		mfp.imr.a = data;
		ChangeInterrupt();
		break;
	 case MFP::IMRB:
		putlog(2, "%s <- $%02x", regname[offset], data);
		mfp.imr.b = data;
		ChangeInterrupt();
		break;
	 case MFP::VR:
		mfp.vr_vec = data & 0xf0;
		mfp.vr_s   = data & 0x08;
		putlog(2, "%s <- $%02x", regname[offset], mfp.GetVR());
		break;
	 case MFP::TACR:
		putlog(2, "%s <- $%02x", regname[offset], data);
		SetTCR(0, data);
		break;
	 case MFP::TBCR:
		putlog(2, "%s <- $%02x", regname[offset], data);
		SetTCR(1, data);
		break;
	 case MFP::TCDCR:
		putlog(2, "%s <- $%02x", regname[offset], data);
		SetTCR(2, data >> 4);
		SetTCR(3, data & 7);
		break;
	 case MFP::TADR:
		putlog(2, "%s <- $%02x", regname[offset], data);
		SetTDR(0, data);
		break;
	 case MFP::TBDR:
		putlog(2, "%s <- $%02x", regname[offset], data);
		SetTDR(1, data);
		break;
	 case MFP::TCDR:
		putlog(2, "%s <- $%02x", regname[offset], data);
		SetTDR(2, data);
		break;
	 case MFP::TDDR:
		putlog(2, "%s <- $%02x", regname[offset], data);
		SetTDR(3, data);
		break;
	 case MFP::SCR:
		putlog(0, "SCR <- $%02x (NOT IMPLEMENTED)", data);
		mfp.scr = data;
		break;
	 case MFP::UCR:
		putlog(2, "%s <- $%02x", regname[offset], data);
		SetUCR(data);
		break;
	 case MFP::RSR:
		putlog(2, "%s <- $%02x", regname[offset], data);
		SetRSR(data);
		break;
	 case MFP::TSR:
		if ((data & 0x0e)) {
			// SO 端子関係の B, H, L ビットは未実装
			putlog(0, "TSR <- $%02x (B,H,L bit NOT IMPLEMENTED)", data);
		}
		mfp.tsr = (mfp.tsr & 0xf0) | (data & 0x0f);
		break;
	 case MFP::UDR:
		if (__predict_false(loglevel >= 2)) {
			if (data == 0x40 || data == 0x41) {
				// MSCTRL だけ頻度が高いのでログレベルを上げておく
				putlog(3, "%s <- $%02x", regname[offset], data);
			} else {
				putlogn("%s <- $%02x", regname[offset], data);
			}
		}
		SetUDR(data);
		break;
	 default:
		break;
	}

	busdata r = wait;
	r |= BusData::Size1;
	return r;
}

busdata
MFPDevice::PeekPort(uint32 offset)
{
	uint8 data;

	switch (offset) {
	 case MFP::GPIP:
		data = mfp.gpip;
		break;
	 case MFP::AER:
		data = mfp.aer;
		break;
	 case MFP::DDR:
		data = mfp.ddr;
		break;
	 case MFP::IERA:
		data = mfp.ier.a;
		break;
	 case MFP::IERB:
		data = mfp.ier.b;
		break;
	 case MFP::IPRA:
		data = mfp.ipr.a;
		break;
	 case MFP::IPRB:
		data = mfp.ipr.b;
		break;
	 case MFP::ISRA:
		data = mfp.isr.a;
		break;
	 case MFP::ISRB:
		data = mfp.isr.b;
		break;
	 case MFP::IMRA:
		data = mfp.imr.a;
		break;
	 case MFP::IMRB:
		data = mfp.imr.b;
		break;
	 case MFP::VR:
		data = mfp.GetVR();
		break;
	 case MFP::TACR:
		data = mfp.timer[0].tcr;
		break;
	 case MFP::TBCR:
		data = mfp.timer[1].tcr;
		break;
	 case MFP::TCDCR:
		data = mfp.GetTCDCR();
		break;
	 case MFP::TADR:
		data = GetTDR(0);
		break;
	 case MFP::TBDR:
		data = GetTDR(1);
		break;
	 case MFP::TCDR:
		data = GetTDR(2);
		break;
	 case MFP::TDDR:
		data = GetTDR(3);
		break;
	 case MFP::SCR:
		data = mfp.scr;
		break;
	 case MFP::UCR:
		data = mfp.ucr;
		break;
	 case MFP::RSR:
		data = mfp.rsr;
		break;
	 case MFP::TSR:
		data = mfp.tsr;
		break;
	 case MFP::UDR:
		data = mfp.udr;
		break;
	 default:
		data = 0xff;
	}
	return data;
}

bool
MFPDevice::PokePort(uint32 offset, uint32 data)
{
	return false;
}

void
MFPDevice::MonitorUpdate(Monitor *, TextScreen& screen)
{
	MFP mfp_;
	std::array<uint8, 4> tdr_;
	int x;
	int y;

	// ローカルにコピー。
	memcpy(&mfp_, &mfp, sizeof(mfp));
	for (int ch = 0; ch < tdr_.size(); ch++) {
		tdr_[ch] = GetTDR(ch);
	}

	screen.Clear();

	// 0         1         2         3         4         5         6
	// 01234567890123456789012345678901234567890123456789012345678901234567890
	// $e88001(GPIP):$00   $e88007(IERA):$00   $e88019(TACR):$00   $e88027
	//                 IER    IMR  IPR  ISR            Mode        TDR Cnt
	// An:MPSC TxEmpty Enable Mask Pend Serv   Timer-A Delay(50us) 255 255

#define A(x)	(baseaddr + (x * 2) + 1)

	// レジスタ生値
	// 1列目(GPIP 関連と VR)
	x = 0;
	y = 0;
	screen.Print(x, y++, "$%06x(GPIP):$%02x", A(MFP::GPIP), mfp_.gpip);
	screen.Print(x, y++, "$%06x(AER): $%02x", A(MFP::AER),  mfp_.aer);
	screen.Print(x, y++, "$%06x(DDR): $%02x", A(MFP::DDR),  mfp_.ddr);
	screen.Print(x, y++, "$%06x(VR):  $%02x", A(MFP::VR),   mfp_.GetVR());
	// 2列目(割り込み関連)
	x = 20;
	y = 0;
	screen.Print(x, y++, "$%06x(IERA):$%02x", A(MFP::IERA), mfp_.ier.a);
	screen.Print(x, y++, "$%06x(IERB):$%02x", A(MFP::IERB), mfp_.ier.b);
	screen.Print(x, y++, "$%06x(IPRA):$%02x", A(MFP::IPRA), mfp_.ipr.a);
	screen.Print(x, y++, "$%06x(IPRB):$%02x", A(MFP::IPRB), mfp_.ipr.b);
	screen.Print(x, y++, "$%06x(ISRA):$%02x", A(MFP::ISRA), mfp_.isr.a);
	screen.Print(x, y++, "$%06x(ISRB):$%02x", A(MFP::ISRB), mfp_.isr.b);
	screen.Print(x, y++, "$%06x(IMRA):$%02x", A(MFP::IMRA), mfp_.imr.a);
	screen.Print(x, y++, "$%06x(IMRB):$%02x", A(MFP::IMRB), mfp_.imr.b);
	// 3列目(タイマー関連)
	x = 40;
	y = 0;
	screen.Print(x, y++, "$%06x(TACR): $%02x", A(MFP::TACR), mfp_.timer[0].tcr);
	screen.Print(x, y++, "$%06x(TBCR): $%02x", A(MFP::TBCR), mfp_.timer[1].tcr);
	screen.Print(x, y++, "$%06x(TCDCR):$%02x", A(MFP::TCDCR), mfp_.GetTCDCR());
	screen.Print(x, y++, "$%06x(TADR): $%02x", A(MFP::TADR), tdr_[0]);
	screen.Print(x, y++, "$%06x(TBDR): $%02x", A(MFP::TBDR), tdr_[1]);
	screen.Print(x, y++, "$%06x(TCDR): $%02x", A(MFP::TCDR), tdr_[2]);
	screen.Print(x, y++, "$%06x(TDDR): $%02x", A(MFP::TDDR), tdr_[3]);
	// 4列目(USART 関連)
	x = 61;
	y = 0;
	screen.Print(x, y++, "$%06x(SCR):$%02x", A(MFP::SCR), mfp_.scr);
	screen.Print(x, y++, "$%06x(UCR):$%02x", A(MFP::UCR), mfp_.ucr);
	screen.Print(x, y++, "$%06x(RSR):$%02x", A(MFP::RSR), mfp_.rsr);
	screen.Print(x, y++, "$%06x(TSR):$%02x", A(MFP::TSR), mfp_.tsr);
	screen.Print(x, y, "$%06x(UDR):", A(MFP::UDR));
	if ((int16)mfp_.udr >= 0) {
		screen.Print(x + 13, y, "$%02x", mfp_.udr);
	} else {
		screen.Print(x + 13, y, "---");
	}

	// 割り込み関連
	y = 9;
	screen.Print(0, y++, "<Interrupt>     %-6s %-4s %-4s %-4s",
		"IER", "IMR", "IPR", "ISR");
	for (int i = 15; i >= 0; i--, y++) {
		uint16 mask = (1 << i);
		screen.Print(0, y, "%c%u:%-12s %-6s %-4s %-4s %-4s",
			i >= 8 ? 'A' : 'B',
			(i % 8),
			intrname[i],
			(mfp_.ier.w & mask) ? "Enable" : "",
			(mfp_.imr.w & mask) ? ""       : "Mask",	// %0 でマスクする
			(mfp_.ipr.w & mask) ? "Pend"   : "",
			(mfp_.isr.w & mask) ? "Serv"   : "");
	}

	// タイマー関連
	x = 40;
	y = 9;
	screen.Print(x, y++, "<Timer> Mode        TDR Next");
	for (int ch = 0; ch < 4; ch++) {
		screen.Print(x, y, "Timer-%c", ch + 'A');
		uint8 tcr = mfp_.timer[ch].tcr;
		if (tcr == 0) {
			screen.Print(x + 8, y, "Stopped");
		} else if (tcr == 8) {
			screen.Print(x + 8, y, "Event");
		} else if (tcr > 8) {
			screen.Print(x + 8, y, "Pulse(%s)", prescale_str[tcr - 8]);
		} else {
			screen.Print(x + 8, y, "Delay(%s)", prescale_str[tcr]);
		}

		screen.Print(x + 20, y, "%3u", tdr_[ch]);
		screen.Print(x + 25, y, "%3u", mfp_.timer[ch].next_tdr & 0xff);
		y++;
	}

	// GPIP 関連
	x = 40;
	y = 9 + 8;
	screen.Puts(x, y++, "<GPIP>    GPIP AER DDR");
	for (int i = 7; i >= 0; i--, y++) {
		uint mask = (1 << i);
		screen.Print(x, y, "b%u %-6s  %u   %u   %s", i, gpipname[i],
			(mfp_.gpip & mask) ? 1 : 0,
			(mfp_.aer & mask) ? 1 : 0,
			(mfp_.ddr & mask) ? "OUT" : "IN");
	}
}

/*static*/ const char *
MFPDevice::intrname[] = {
	"RTC Alarm",
	"EXPON",
	"POW SW",
	"FM IRQ",
	"Timer-D",
	"Timer-C",
	"V-Disp",
	"---",
	"Timer-B",
	"MPSC TxErr",
	"MPSC TxEmpty",
	"MPSC RxErr",
	"MPSC RxFull",
	"Timer-A",
	"CRTC IRQ",
	"H-Sync",
};

/*static*/ const char *
MFPDevice::gpipname[] = {
	"ALARM",
	"EXPON",
	"POW SW",
	"FMIRQ",
	"V-Disp",
	"------",
	"CIRQ",
	"H-Sync",
};

// TxCR の設定値から分周期間というかメインカウンタ1回分の時間を返す
/*static*/ const uint
MFPDevice::prescale_ns[8] = {
	0,
	1000,	// /4
	2500,	// /10
	4000,	// /16
	12500,	// /50
	16000,	// /64
	25000,	// /100
	50000,	// /200
};

/*static*/ const char *
MFPDevice::prescale_str[8] = {
	"",
	"1us",
	"2.5us",
	"4us",
	"12.5us",
	"16us",
	"25us",
	"50us",
};

// タイマーのチャンネルから MFP 割り込みチャンネル番号に変換する
/*static*/ const uint
MFPDevice::timer_vector[4] = {
	MFP::INTR_TIMER_A,
	MFP::INTR_TIMER_B,
	MFP::INTR_TIMER_C,
	MFP::INTR_TIMER_D,
};

// IER[AB] に data をセットする
void
MFPDevice::SetIER(uint8& ier, uint8& ipr, uint32 data)
{
	// clr は、data によって %1 -> %0 に変化するビットだけを立てたもの。
	uint8 clr = ier & (ier ^ data);

	// IER を下げたところは IPR も下げる。(紙資料 p4-4)
	ipr &= ~clr;

	// どちらにしても新しい IER によって信号線の状態は変える。
	ier = data;
	ChangeInterrupt();
}

// TxCR をセットする
void
MFPDevice::SetTCR(int ch, uint32 data)
{
	uint32 prev;

	prev = mfp.timer[ch].tcr;
	data &= 0x0f;

	mfp.timer[ch].tcr = data;
	if (data == 0) {
		if (prev == 0) {
			// 0 -> 0 なら何も起きない
			return;
		} else if (prev < 8) {
			// タイマー停止
			putlog(1, "Timer-%c Stop", ch + 'A');

			// ここで TDR を確定させる。
			// タイマー動作中の TDR の参照はイベント終了時刻から現在値を逆算
			// していたが、イベントを停止すると求められなくなるので、イベント
			// 停止前に求めておく。
			uint64 period = prescale_ns[prev];
			uint64 now = scheduler->GetVirtTime();
			mfp.timer[ch].tdr = ((event[ch].vtime - now) + period - 1) / period;

			// でイベント停止
			scheduler->StopEvent(event[ch]);
			return;
		} else if (prev == 8) {
			// イベントカウントモードから停止しても何も起きないはず
		} else {
			// パルス幅測定モード (未実装)
		}
		return;

	} else if (data < 8) {
		// タイマー開始

		// TDR は触らずイベントを開始。
		uint64 period = prescale_ns[mfp.timer[ch].tcr];
		event[ch].time = period * mfp.timer[ch].tdr;
		scheduler->RestartEvent(event[ch]);

		putlog(1, "Timer-%c Start Delay mode(%u x %u.%uusec)",
			ch + 'A',
			mfp.timer[ch].tdr,
			(uint)period / 1000,
			(uint)(period / 100) % 10);
		return;

	} else if (data == 8) {
		// イベントカウントモード

		// TDR をロードする
		mfp.timer[ch].next_tdr = mfp.timer[ch].tdr;
		putlog(1, "Timer-%c Event count mode", ch + 'A');
		return;

	} else {
		putlog(0, "T%cCR Pulse-Width Measurement Mode $%02x (NOT IMPLEMENTED)",
			ch + 'A', data);
		return;
	}
}

// TxDR の読み出し
// (副作用はないので Peek 系からも呼び出してよい)
uint8
MFPDevice::GetTDR(int ch) const
{
	uint8 tcr = mfp.timer[ch].tcr;

	if (tcr == 0) {
		// 停止中なら tdr に現在値が保持されている
		return mfp.timer[ch].tdr & 0xff;

	} else if (tcr < 8) {
		// ディレイモード動作中なら現在値は保持していないので、
		// プリスケーラと終了時刻から算出。

		uint64 period = prescale_ns[tcr];
		uint64 now = scheduler->GetVirtTime();
		uint8 data = ((event[ch].vtime - now) + period - 1) / period;
		return data;

	} else {
		// イベントモード、パルス幅測定モードは未対応
		return mfp.timer[ch].tdr & 0xff;

	}
}

// TxDR への書き込み
void
MFPDevice::SetTDR(int ch, uint32 data)
{
	uint8& tcr = mfp.timer[ch].tcr;

	if (data == 0) {
		data = 256;
	}
	mfp.timer[ch].next_tdr = data;

	if (tcr == 0) {
		// 停止中の書き込みはメインカウンタも更新。
		mfp.timer[ch].tdr = data;

	} else if (tcr < 8) {
		// ディレイモード動作中の書き込みは
		// メインカウンタは影響を受けない。

	} else {
		// イベントモード、パルス幅測定モードは未対応
		putlog(0, "Write T%cDR $%02x during unsupported mode (NOT IMPLEMENTED)",
			ch + 'A', mfp.timer[ch].next_tdr);
	}
}

// イベントコールバック
void
MFPDevice::TimerCallback(Event& ev)
{
	int& ch = ev.code;

	// TDR がゼロになったところで呼ばれるので、TDR をリロードする
	mfp.timer[ch].tdr = mfp.timer[ch].next_tdr;

	// 割り込みを上げる
	SetInterrupt(timer_vector[ch]);

	// 次のイベントをセット
	// XXX 1回のイベントで複数のタイムアウトパルスを跨ぐ状況は考慮してない

	// period が1メインパルス期間
	// time が1タイムアウトパルス期間
	uint64 period = prescale_ns[mfp.timer[ch].tcr];
	ev.time = period * mfp.timer[ch].tdr;
	scheduler->RestartEvent(ev);
}

// HSync 入力
void
MFPDevice::SetHSync(bool hsync)
{
	SetGPIP(MFP::GPIP_HSYNC, hsync);
}

// VDisp 入力
void
MFPDevice::SetVDisp(bool vdisp)
{
	// TAI への入力はイベントカウントモードの時のみ
	if (mfp.timer[0].tcr == 8) {
		// AER が示す方向と同じエッジ方向の時
		bool aer = (mfp.aer & (1 << MFP::GPIP_VDISP));
		if (aer == vdisp) {
			// カウンタを減算
			mfp.timer[0].tdr--;
			if (mfp.timer[0].tdr == 0) {
				// 0 になったらリロード
				mfp.timer[0].tdr = mfp.timer[0].next_tdr;

				SetInterrupt(MFP::INTR_TIMER_A);
			}
		}
	}

	// GPIP への入力
	SetGPIP(MFP::GPIP_VDISP, vdisp);
}

// 電源ボタン状態入力 (PowerDevice から呼ばれる)
void
MFPDevice::SetPowSW(bool powsw)
{
	SetGPIP(MFP::GPIP_POWSW, powsw);
}

// ALARM 信号入力 (RTC の ALARM_OUT 出力端子)
void
MFPDevice::SetAlarmOut(bool alarm)
{
	SetGPIP(MFP::GPIP_ALARM, alarm);
}

// GPIP への入力。
// num はビット番号(0..7)。
void
MFPDevice::SetGPIP(int num, bool val)
{
	bool aer = (mfp.aer & (1 << num));
	bool old = (mfp.gpip & (1 << num));

	// 立ち上がりか立ち下がりで割り込みを上げる (AER による)
	// AER OLD VAL   Interrupt
	//   0   0   0   0
	//   0   0   1   0
	//   0   1   0   1
	//   0   1   1   0
	//   1   0   0   0
	//   1   0   1   1
	//   1   1   0   0
	//   1   1   1   0
	if ((old != val) && (aer == val)) {
		// ここで割り込みを上げる
		// XXX GPIP ビット番号と INTR の対応どうなってんの
		int intr;
		if (num < 4) {
			intr = num;
		} else if (num < 6) {
			intr = num + 2;
		} else {
			intr = num - MFP::GPIP_CIRQ + MFP::INTR_CRTC_IRQ;
		}
		SetInterrupt(intr);
	}

	if (val) {
		mfp.gpip |= (1 << num);
	} else {
		mfp.gpip &= ~(1 << num);
	}
}

// UCR への書き込み
void
MFPDevice::SetUCR(uint32 data)
{
	mfp.ucr = data;

	if ((mfp.ucr & MFP::UCR_CLK) == 0) {
		putlog(0, "UCR CLK=0 (NOT IMPLEMENTED)");
	}
	if (((mfp.ucr & MFP::UCR_WL) >> 5) != 0) {
		putlog(0, "UCR WL=%u (NOT IMPLEMENTED)", (mfp.ucr >> 5) & 3);
	}
	if (((mfp.ucr & MFP::UCR_ST) >> 3) != 1) {
		putlog(0, "UCR ST=%u (NOT IMPLEMENTED)", (mfp.ucr >> 3) & 3);
	}
}

// RSR への書き込み
void
MFPDevice::SetRSR(uint32 data)
{
	uint32 old = mfp.rsr;

	mfp.rsr = data;

	if ((old & MFP::RSR_RE) == 0 && (mfp.rsr & MFP::RSR_RE)) {
		// RE 0 -> 1
	}
	if ((old & MFP::RSR_RE) && (mfp.rsr & MFP::RSR_RE) == 0) {
		// RE 1 -> 0
		// 上位4ビットのステータスを %0 に
		mfp.rsr &= ~(MFP::RSR_OE | MFP::RSR_PE | MFP::RSR_FE);
		ClearBF();
	}
}

// RSR::BF (Buffer Full) をセットする
void
MFPDevice::SetBF()
{
	mfp.rsr |= MFP::RSR_BF;
}

// RSR::BF (Buffer Full) をクリアする
void
MFPDevice::ClearBF()
{
	mfp.rsr &= ~MFP::RSR_BF;

	// バッファが空いたのでキーボードに次の転送を催促。
	// 実際のハードウェアでは、キーボードが本体 MFP が Ready になるのを待って
	// いるが、それに相当。
	// システムポートの KEYCTRL はキーボード側で処理している。
	if (RxIsReady()) {
		keyboard->Ready();
	}
}

// UDR への書き込み
void
MFPDevice::SetUDR(uint32 data)
{
	if ((mfp.tsr & MFP::TSR_TE)) {
		keyboard->Command(data);
	}
}

// __
// RR 線の状態を返す。アサートなら true。
bool
MFPDevice::IsRR() const
{
	//        __
	// 実際の RR 信号線は、BF が立っていて、PE | FE がクリアされている場合に
	// アサートされる。ただしエミュレータでは Parity Error も Frame Error も
	// 起きないので、この2ビットは無視する。
	return ((mfp.rsr & MFP::RSR_BF) != 0);
}

// キーボードからの受信が可能なら true を返す。
bool
MFPDevice::RxIsReady() const
{
	// __
	// RR は本来 MFP から CPU/DMAC など上位に対しての Ready で、
	// UDR にデータが用意できたので引き取ってほしいの意。
	// X680x0 ではこれを CPU/DMAC (上位) 向けではなく、下位のキーボードからの
	// フロー制御に流用しており、
	// 上位への(本来の)「UDR にデータが用意できた」(IsRR() = true) は、
	// キーボードに対しては「UDR にデータがあるので送信禁止」、
	// 上位への(本来の)「UDR にデータが用意できていない」(IsRR() = false) は、
	// キーボードに対しては「UDR は空いているので送信してよい」と
	// ここで意味が論理反転することに注意。
	//
	// また、システムポートの KEYCTRL は X680x0Keyboard 側で処理しているので
	// こちらでは不要。
	//
	// UDR が空いていてキーボードからデータが送られてきている間は RR は Ready
	// だが、そもそもキーボードは送出作業中で、新たな送信はできないはずなので
	// この場合、つまり key_event の動作中も false (送信不可) にする。

	return !IsRR() && !key_event.IsRunning();
}

// キーボードからデータを受信
void
MFPDevice::Rx(uint32 data)
{
	// ここは受信したビットをシフトレジスタへ入れ始めたところに相当する
	// ので、まだ Buffer Full は立てない。

	key_event.code = data;
	scheduler->RestartEvent(key_event);
}

// キーボードからのデータが受信し終わった頃に呼ばれるイベント。
void
MFPDevice::KeyCallback(Event& ev)
{
	// UDR にデータを置いて Buffer Full をセット
	mfp.udr = ev.code;
	SetBF();

	// Manual 7-5 7.2.1 Receiver Interrupt Channels より
	// | 割り込みは1文字受信ごとに1回しか発生しませんが、2つの専用割り込み
	// | チャネルにより、正常な受信状態と異常な受信状態に別々のベクター番号を
	// | 割り当てることができます。受信したワードにエラーがあり、エラー割り込み
	// | チャネルが有効な場合は、エラーチャネルのみに割り込みが発生します。
	// | しかし、エラーチャネルが無効の場合、エラー状態の割り込みは、
	// | バッファフル状態の割り込みと一緒にバッファフル割り込みチャネルに
	// | 生成されます。どのエラー条件で割り込みが発生したかを判断するためには、
	// | 常にRSRを読み出す必要があります。
	// XXX: まだエラー状態は実装されていない、なおエラーは起きないので
	// メモだけのままかも知れない。

	SetInterrupt(MFP::INTR_RXFULL);
}

// 必要なら割り込みを上げる。
// ch は MFP 内の割り込みチャンネル番号。MFP::INTR_*
void
MFPDevice::SetInterrupt(uint ch)
{
	assertmsg(ch < 16, "ch=%u", ch);
	__assume(ch < 16);

	uint b = (1U << ch);

	// IER で割り込みが無効なら何もしない
	if ((mfp.ier.w & b) == 0) {
		return;
	}

	// IPR にペンディングを立てる
	mfp.ipr.w |= b;

	// 割り込み信号線の状態を変える
	ChangeInterrupt();
}

// 割り込み信号線の状態を変える
void
MFPDevice::ChangeInterrupt()
{
	// ペンディングの割り込みがあればアサートされる、
	// IPR をクリアするか IMR をクリアすると /IRQ はネゲートされる。
	bool irq = ((mfp.ipr.w & mfp.imr.w) != 0);
	interrupt->ChangeINT(this, irq);
}

// 割り込みアクノリッジ
busdata
MFPDevice::InterruptAcknowledge()
{
	busdata data;

	// IPR が誰も立ってなければ、無視するっぽい?
	// 実際には MFP がバスエラーを出しているのではなく、MFP は俺知らない
	// と思って放置してると上位回路 (たぶん SAKI ちゃん) がタイムアウトして
	// MPU にバスエラー応答する。
	uint32 pending = mfp.ipr.w;
	if (__predict_false(pending == 0)) {
		data.SetBusErr();
	} else {
		// ch はこの時点で IPR に立ってる一番優先度の高いチャンネルのビット
		int ch = 31 - __builtin_clz(pending);
		uint b = (1U << ch);

		// MPU にベクタを引き渡したので Pending をクリア
		mfp.ipr.w &= ~b;
		ChangeInterrupt();

		// ソフトウェア EOI なら In Service を立てる
		if (mfp.vr_s) {
			mfp.isr.w |= b;
		}

		data = mfp.vr_vec + ch;
	}

	// /IEI ネゲート時の /IACK と /IEI の関係がいまいち分からないけど、
	// とりあえず /IACK がアサートされてから 3 MFP クロック後に
	// データバスにベクタを供給、次の 1 MFP クロックで DTACK 応答。
	// で次の 1 MFP クロック以内に CPU が /IACK を下げるとして、
	// 計 5 MFP クロック (= 1.25_usec) としておく。
	const busdata intack_wait = busdata::Wait(5 * 250_nsec);
	data |= intack_wait;

	return data;
}

/*static*/ const char *
MFPDevice::regname[MFP::RegMax] = {
	"GPIP",
	"AER",
	"DDR",
	"IERA",
	"IERB",
	"IPRA",
	"IPRB",
	"ISRA",
	"ISRB",
	"IMRA",
	"IMRB",
	"VR",
	"TACR",
	"TBCR",
	"TCDCR",
	"TADR",
	"TBDR",
	"TCDR",
	"TDDR",
	"SCR",
	"UCR",
	"RSR",
	"TSR",
	"UDR",
};
