//
// nono
// Copyright (C) 2020 nono project
// Licensed under nono-license.txt
//

//
// LUNA システムクロック
//

// システムクロックは実時間または仮想時間のいずれかに同期するため、
// システムクロック独自の時刻 stime を持つ。
// stime は、割り込みを発生させるごとに、LUNA-I なら 16.67msec、
// LUNA-88K なら 10msec 進行する。

#include "sysclk.h"
#include "config.h"
#include "interrupt.h"
#include "monitor.h"
#include "rtc.h"
#include "scheduler.h"
#include "syncer.h"

// コンストラクタ
SysClkDevice::SysClkDevice()
	: inherited(OBJ_SYSCLK)
{
	monitor = gMonitorManager->Regist(ID_MONITOR_SYSCLK, this);
	monitor->func = ToMonitorCallback(&SysClkDevice::MonitorUpdate);
	monitor->SetSize(31, 3);
}

// デストラクタ
SysClkDevice::~SysClkDevice()
{
}

// 初期化
bool
SysClkDevice::Init()
{
	if (inherited::Init() == false) {
		return false;
	}

	interrupt = GetInterruptDevice();

	// クロックは機種ごとに異なる
	const ConfigItem& item = gConfig->Find("!sysclk-freq");
	freq = item.AsInt();
	period = 1_sec / freq;
	event.time = period;

	// クロックの同期方法
	sync_rt = GetRTCDevice()->GetSyncRealtime();

	event.func = ToEventCallback(&SysClkDevice::Callback);
	event.code = 0;
	event.SetName("System Clock");
	scheduler->RegistEvent(event);

	return true;
}

// 電源オン/リセット
void
SysClkDevice::ResetHard(bool poweron)
{
	// XXX たぶん電源オンだけで作用する?
	if (poweron) {
		if (sync_rt) {
			stime = GetSyncer()->GetRealTime();
		}

		// 電源オンですぐ起動
		scheduler->RestartEvent(event);
	}

	// さすがに割り込みネゲートくらいするよな?
	sysint = false;
	ChangeInterrupt();
}

busdata
SysClkDevice::ReadPort(uint32 offset)
{
	// 1 ポートしかない
	busdata data = GetInt();
	data |= BusData::Size1;
	return data;
}

busdata
SysClkDevice::WritePort(uint32 offset, uint32 data)
{
	// 何か書き込まれたら、フラグをクリア
	putlog(3, "Clear system clock interrupt");
	sysint = false;
	ChangeInterrupt();

	busdata r = BusData::Size1;
	return r;
}

busdata
SysClkDevice::PeekPort(uint32 offset)
{
	return GetInt();
}

bool
SysClkDevice::PokePort(uint32 offset, uint32 data)
{
	return false;
}

void
SysClkDevice::MonitorUpdate(Monitor *, TextScreen& screen)
{
	int y;

	screen.Clear();
	y = 0;

	int width = stime < 86400_sec ? 20 : 25;
	screen.Print(0, y++, "Freq: %u Hz", freq);
	screen.Print(0, y++, "Sync: %s", sync_rt ? "Real Time" : "Virtual Time");
	screen.Print(0, y++, "Time: %*s", width, TimeToStr(stime).c_str());
}

// 割り込み状態を取得(?)
uint8
SysClkDevice::GetInt() const
{
	uint8 data = 0;

	if (sysint) {
		data |= 0x80;
	}
	return data;
}

// イベントハンドラ
void
SysClkDevice::Callback(Event& ev)
{
	putlog(3, "Interrupt occured");

	// システムクロックとしての現在時刻。
	// 足す前は前回の割り込み発生時刻なので period を足すと現在時刻。
	stime += period;

	sysint = true;
	ChangeInterrupt();

	if (sync_rt) {
		scheduler->StartRealtimeEvent(ev, stime, period);
	} else {
		scheduler->StartEvent(ev);
	}
}

// 割り込み信号線の状態を変える。
void
SysClkDevice::ChangeInterrupt()
{
	interrupt->ChangeINT(this, sysint);
}
